"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createCoreMutationsProvider = void 0;
const chalk_1 = __importDefault(require("chalk"));
const os_1 = require("os");
const pluralize_1 = require("../../output/pluralize");
const LazyCache_1 = require("../../services/LazyCache");
const FileInfoCache_1 = require("../../shared/FileInfoCache");
const maps_1 = require("../../shared/maps");
const NameGenerator_1 = require("../../shared/NameGenerator");
const collectFilteredNodes_1 = require("../collectFilteredNodes");
const createFileNamesAndServices_1 = require("../createFileNamesAndServices");
const findMutationsInFile_1 = require("../findMutationsInFile");
const WaveTracker_1 = require("./tracking/WaveTracker");
/**
 * Creates a mutations provider that runs the core mutations within TypeStat.
 *
 * @param allModifiedFileNames   Set to mark names of all files that were modified.
 * @returns Provider to run the core mutations, if needed.
 */
const createCoreMutationsProvider = (allModifiedFiles) => {
    return (options) => {
        const enabledFixes = Object.keys(options.fixes).filter((key) => options.fixes[key]);
        if (!enabledFixes) {
            return undefined;
        }
        const fileNamesAndServicesCache = createFileNamesAndServicesCache(options);
        const waveTracker = new WaveTracker_1.WaveTracker();
        let lastFileIndex = -1;
        let waveIndex = 1;
        return async () => {
            const startTime = Date.now();
            const fileMutationsByFileName = new Map();
            const { fileNames, services } = fileNamesAndServicesCache.get();
            const waveStartedFromBeginning = lastFileIndex <= 0;
            let addedMutations = 0;
            if (waveIndex === 1) {
                options.output.stdout([
                    chalk_1.default.blueBright(`Starting the core mutation engine.`),
                    chalk_1.default.blue(` This terminal will log whenever a "wave" of mutations are written to files.`),
                ].join(""));
                options.output.stdout(chalk_1.default.gray(`Core mutations will complete when two waves pass with no mutations.`));
                options.output.stdout(chalk_1.default.gray(`The following ${pluralize_1.pluralize(enabledFixes.length, "fix", "es")} will be applied:`));
                for (let i = 0; i < enabledFixes.length; i += 1) {
                    options.output.stdout(chalk_1.default.gray(`\t* ${enabledFixes[i]}${i === enabledFixes.length - 1 ? os_1.EOL : ""}`));
                }
            }
            options.output.stdout(chalk_1.default.gray(`Starting wave ${waveIndex}...`));
            for (lastFileIndex += 1; lastFileIndex < fileNames.length; lastFileIndex += 1) {
                const fileName = fileNames[lastFileIndex];
                const sourceFile = services.program.getSourceFile(fileName);
                if (sourceFile === undefined) {
                    options.output.stderr(`Could not find TypeScript source file for '${fileName}'.`);
                    continue;
                }
                const filteredNodes = collectFilteredNodes_1.collectFilteredNodes(options, sourceFile);
                const foundMutations = await findMutationsInFile_1.findMutationsInFile({
                    fileInfoCache: new FileInfoCache_1.FileInfoCache(filteredNodes, services, sourceFile),
                    filteredNodes,
                    nameGenerator: new NameGenerator_1.NameGenerator(sourceFile.fileName),
                    options,
                    services,
                    sourceFile,
                });
                if (foundMutations !== undefined && foundMutations.length !== 0) {
                    addedMutations += foundMutations.length;
                    fileMutationsByFileName.set(fileName, foundMutations);
                }
                if (addedMutations > 100 || (addedMutations !== 0 && Date.now() - startTime > 10000)) {
                    break;
                }
            }
            if (lastFileIndex === fileNames.length) {
                lastFileIndex = -1;
                // Only recreate the language service once we've visited every file
                // This way we don't constantly re-scan many of the source files each wave
                // Eventually it would be nice to support incremental updates
                // See https://github.com/JoshuaKGoldberg/TypeStat/issues/36
                fileNamesAndServicesCache.clear();
            }
            for (const fileName of fileMutationsByFileName.keys()) {
                allModifiedFiles.add(fileName);
            }
            const fileMutations = waveStartedFromBeginning && fileMutationsByFileName.size === 0 ? undefined : maps_1.convertMapToObject(fileMutationsByFileName);
            const mutationsCount = fileMutations ? Object.keys(fileMutations).length : 0;
            options.output.stdout(chalk_1.default.gray(`Completed wave ${waveIndex}. Wrote mutations to ${mutationsCount} ${pluralize_1.pluralize(mutationsCount, "file")}.`));
            if (!fileMutations) {
                options.output.stdout(chalk_1.default.blueBright(`Done.${os_1.EOL}`));
                return undefined;
            }
            waveIndex += 1;
            if (waveTracker.addAndCheck(fileMutations)) {
                options.output.stdout(chalk_1.default.redBright(`It looks like TypeStat has ended up in an infinite loop.${os_1.EOL}`));
                options.output.stdout(chalk_1.default.red(`Bailing out of applying more mutations.${os_1.EOL}`));
                options.output.stdout(chalk_1.default.red(`Please file an issue to help us fix the bug for you:${os_1.EOL}`));
                options.output.stdout(chalk_1.default.redBright(`https://github.com/JoshuaKGoldberg/TypeStat${os_1.EOL}`));
                return undefined;
            }
            return { mutationsWave: { fileMutations } };
        };
    };
};
exports.createCoreMutationsProvider = createCoreMutationsProvider;
const createFileNamesAndServicesCache = (options) => {
    return new LazyCache_1.LazyCache(() => {
        options.output.log?.("Preparing language services to visit files...");
        const { fileNames, services } = createFileNamesAndServices_1.createFileNamesAndServices(options);
        options.output.log?.(`Prepared language services for ${fileNames.length} files...`);
        return { fileNames, services };
    });
};
//# sourceMappingURL=createCoreMutationsProvider.js.map