"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createFileRenamesProvider = void 0;
const chalk_1 = __importDefault(require("chalk"));
const mz_1 = require("mz");
const pluralize_1 = require("../../../output/pluralize");
const createSingleUseProvider_1 = require("../../createSingleUseProvider");
const getNewFileName_1 = require("./getNewFileName");
/**
 * Creates a mutations provider that renames files from JavaScript to TypeScript.
 *
 * @param allModifiedFileNames   Set to mark names of all files that were modified.
 * @returns Provider to rename files from JavaScript to TypeScript, if needed.
 */
const createFileRenamesProvider = (allModifiedFiles) => {
    return createSingleUseProvider_1.createSingleUseProvider("Renaming files from JavaScript to TypeScript", (options) => {
        // If the options don't specify to rename extension, only make sure no JS files are included
        if (!options.files.renameExtensions) {
            const jsFileNames = options.fileNames.filter(fileNameIsJavaScript);
            if (jsFileNames.length === 0) {
                return undefined;
            }
            return async () => {
                throw new Error([
                    "The following JavaScript files were included in the project but files.renameExtensions is not enabled.",
                    "TypeStat does not yet support annotating JavaScript files.",
                    ...jsFileNames.map((fileName) => `\t${fileName}`),
                    "See https://github.com/JoshuaKGoldberg/TypeStat/blob/main/docs/Files.md for details.",
                ].join("\n"));
            };
        }
        return async () => {
            const newFileNames = new Set(options.fileNames);
            const filesToRename = (await Promise.all(options.fileNames.filter(fileNameIsJavaScript).map(async (fileName) => ({
                newFileName: await getNewFileName_1.getNewFileName(options.files.renameExtensions, fileName, async (filePath) => (await mz_1.fs.readFile(filePath)).toString()),
                oldFileName: fileName,
            })))).filter(({ newFileName, oldFileName }) => newFileName !== oldFileName);
            if (filesToRename.length !== 0) {
                for (const { oldFileName, newFileName } of filesToRename) {
                    await mz_1.fs.rename(oldFileName, newFileName);
                    newFileNames.delete(oldFileName);
                    newFileNames.add(newFileName);
                    allModifiedFiles.add(newFileName);
                }
            }
            const renamedFileNames = Array.from(newFileNames);
            options.output.stdout(chalk_1.default.gray(`Renamed ${renamedFileNames.length} ${pluralize_1.pluralize(renamedFileNames.length, "file")} from JavaScript to TypeScript.`));
            return {
                newOptions: { ...options, fileNames: renamedFileNames },
            };
        };
    });
};
exports.createFileRenamesProvider = createFileRenamesProvider;
const javaScriptExtensionMatcher = /\.(c|m)?jsx?/i;
const fileNameIsJavaScript = (fileName) => javaScriptExtensionMatcher.test(fileName);
//# sourceMappingURL=index.js.map