"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createInstallMissingTypesProvider = void 0;
const builtin_modules_1 = __importDefault(require("builtin-modules"));
const sets_1 = require("../../shared/sets");
const createFileNamesAndServices_1 = require("../createFileNamesAndServices");
const createSingleUseProvider_1 = require("../createSingleUseProvider");
const collectExistingTypingPackages_1 = require("./missingTypes/collectExistingTypingPackages");
const collectPackageManagerRunner_1 = require("./missingTypes/collectPackageManagerRunner");
const collectReferencedPackageNames_1 = require("./missingTypes/collectReferencedPackageNames");
const filterTypedPackageNames_1 = require("./missingTypes/filterTypedPackageNames");
const uniqueBuiltinModules = new Set(builtin_modules_1.default);
/**
 * Creates a mutations provider that installs missing types modules.
 *
 * @returns Provider to install missing types modules, if needed.
 */
const createInstallMissingTypesProvider = () => {
    return createSingleUseProvider_1.createSingleUseProvider("Installing missing @types modules", (options) => {
        const { missingTypes } = options.package;
        if (missingTypes === undefined) {
            return undefined;
        }
        return async () => {
            await (async () => {
                // Collect package names already present in the package file
                const existingPackageNames = await collectExistingTypingPackages_1.collectExistingTypingPackages(options, options.package.file);
                if (existingPackageNames === undefined) {
                    return;
                }
                // Collect every package name referenced by every file in the project
                const { services } = createFileNamesAndServices_1.createFileNamesAndServices(options);
                const referencedPackageNames = collectReferencedPackageNames_1.collectReferencedPackageNames(services);
                // Ignore package names already referenced in package.json or that don't exist in DefinitelyTyped
                const missingPackageNames = sets_1.setSubtract(referencedPackageNames, new Set(existingPackageNames), uniqueBuiltinModules);
                const missingTypedPackageNames = await filterTypedPackageNames_1.filterTypedPackageNames(Array.from(missingPackageNames));
                if (missingTypedPackageNames.length === 0) {
                    return;
                }
                // Run the installation command using the requested or detected package manager
                const packageManagerRunner = await collectPackageManagerRunner_1.collectPackageManagerRunner(options, missingTypes);
                await packageManagerRunner(options, missingTypedPackageNames);
            })();
            return {
                mutationsWave: {},
            };
        };
    });
};
exports.createInstallMissingTypesProvider = createInstallMissingTypesProvider;
//# sourceMappingURL=createInstallMissingTypesProvider.js.map